@@ -133,6 +133,8 @@ typedef enum {
   C_HTTP2_WINDOW_UPDATE,
   C_HTTP3,
   C_HTTP3_ONLY,
+  C_HTTP3_PSEUDO_HEADERS_ORDER,
+  C_HTTP3_SETTINGS,
   C_IGNORE_CONTENT_LENGTH,
   C_IMPERSONATE,
   C_INCLUDE,
@@ -227,6 +229,7 @@ typedef enum {
   C_PROXY1_0,
   C_PROXYTUNNEL,
   C_PUBKEY,
+  C_QUIC_TRANSPORT_PARAMETERS,
   C_QUOTE,
   C_RANDOM_FILE,
   C_RANGE,
