@@ -57,27 +57,29 @@ int nghttp3_stream_new(nghttp3_stream **pstream, int64_t stream_id,
     return NGHTTP3_ERR_NOMEM;
   }
 
-  memset(stream, 0, sizeof(*stream));
-
-  stream->out_chunk_objalloc = out_chunk_objalloc;
-  stream->stream_objalloc = stream_objalloc;
+  *stream = (nghttp3_stream){
+    .out_chunk_objalloc = out_chunk_objalloc,
+    .stream_objalloc = stream_objalloc,
+    .qpack_blocked_pe.index = NGHTTP3_PQ_BAD_INDEX,
+    .mem = mem,
+    .rx =
+      {
+        .http.status_code = -1,
+        .http.content_length = -1,
+        .http.pri.urgency = NGHTTP3_DEFAULT_URGENCY,
+      },
+    .error_code = NGHTTP3_H3_NO_ERROR,
+  };
 
   nghttp3_tnode_init(&stream->node, stream_id);
 
-  nghttp3_ringbuf_init(&stream->frq, 0, sizeof(nghttp3_frame_entry), mem);
+  nghttp3_ringbuf_init(&stream->frq, 0, sizeof(nghttp3_frame), mem);
   nghttp3_ringbuf_init(&stream->chunks, 0, sizeof(nghttp3_buf), mem);
   nghttp3_ringbuf_init(&stream->outq, 0, sizeof(nghttp3_typed_buf), mem);
   nghttp3_ringbuf_init(&stream->inq, 0, sizeof(nghttp3_buf), mem);
 
   nghttp3_qpack_stream_context_init(&stream->qpack_sctx, stream_id, mem);
 
-  stream->qpack_blocked_pe.index = NGHTTP3_PQ_BAD_INDEX;
-  stream->mem = mem;
-  stream->rx.http.status_code = -1;
-  stream->rx.http.content_length = -1;
-  stream->rx.http.pri.urgency = NGHTTP3_DEFAULT_URGENCY;
-  stream->error_code = NGHTTP3_H3_NO_ERROR;
-
   if (callbacks) {
     stream->callbacks = *callbacks;
   }
@@ -123,8 +125,7 @@ static void delete_out_chunks(nghttp3_ringbuf *chunks,
     buf = nghttp3_ringbuf_get(chunks, i);
 
     if (nghttp3_buf_cap(buf) == NGHTTP3_STREAM_MIN_CHUNK_SIZE) {
-      nghttp3_objalloc_chunk_release(out_chunk_objalloc,
-                                     (nghttp3_chunk *)(void *)buf->begin);
+      nghttp3_objalloc_chunk_release(out_chunk_objalloc, (void *)buf->begin);
       continue;
     }
 
@@ -135,17 +136,17 @@ static void delete_out_chunks(nghttp3_ringbuf *chunks,
 }
 
 static void delete_frq(nghttp3_ringbuf *frq, const nghttp3_mem *mem) {
-  nghttp3_frame_entry *frent;
+  nghttp3_frame *fr;
   size_t i, len = nghttp3_ringbuf_len(frq);
 
   for (i = 0; i < len; ++i) {
-    frent = nghttp3_ringbuf_get(frq, i);
-    switch (frent->fr.hd.type) {
+    fr = nghttp3_ringbuf_get(frq, i);
+    switch (fr->hd.type) {
     case NGHTTP3_FRAME_HEADERS:
-      nghttp3_frame_headers_free(&frent->fr.headers, mem);
+      nghttp3_frame_headers_free(&fr->headers, mem);
       break;
     case NGHTTP3_FRAME_PRIORITY_UPDATE:
-      nghttp3_frame_priority_update_free(&frent->fr.priority_update, mem);
+      nghttp3_frame_priority_update_free(&fr->priority_update, mem);
       break;
     default:
       break;
@@ -222,10 +223,9 @@ nghttp3_ssize nghttp3_read_varint(nghttp3_varint_read_state *rvint,
   return (nghttp3_ssize)len;
 }
 
-int nghttp3_stream_frq_add(nghttp3_stream *stream,
-                           const nghttp3_frame_entry *frent) {
+int nghttp3_stream_frq_add(nghttp3_stream *stream, const nghttp3_frame *fr) {
   nghttp3_ringbuf *frq = &stream->frq;
-  nghttp3_frame_entry *dest;
+  nghttp3_frame *dest;
   int rv;
 
   if (nghttp3_ringbuf_full(frq)) {
@@ -238,37 +238,37 @@ int nghttp3_stream_frq_add(nghttp3_stream *stream,
   }
 
   dest = nghttp3_ringbuf_push_back(frq);
-  *dest = *frent;
+  *dest = *fr;
 
   return 0;
 }
 
 int nghttp3_stream_fill_outq(nghttp3_stream *stream) {
   nghttp3_ringbuf *frq = &stream->frq;
-  nghttp3_frame_entry *frent;
+  nghttp3_frame *fr;
   int data_eof;
   int rv;
 
   for (; nghttp3_ringbuf_len(frq) &&
          stream->unsent_bytes < NGHTTP3_MIN_UNSENT_BYTES;) {
-    frent = nghttp3_ringbuf_get(frq, 0);
+    fr = nghttp3_ringbuf_get(frq, 0);
 
-    switch (frent->fr.hd.type) {
+    switch (fr->hd.type) {
     case NGHTTP3_FRAME_SETTINGS:
-      rv = nghttp3_stream_write_settings(stream, frent);
+      rv = nghttp3_stream_write_settings(stream, &fr->settings);
       if (rv != 0) {
         return rv;
       }
       break;
     case NGHTTP3_FRAME_HEADERS:
-      rv = nghttp3_stream_write_headers(stream, frent);
+      rv = nghttp3_stream_write_headers(stream, &fr->headers);
       if (rv != 0) {
         return rv;
       }
-      nghttp3_frame_headers_free(&frent->fr.headers, stream->mem);
+      nghttp3_frame_headers_free(&fr->headers, stream->mem);
       break;
     case NGHTTP3_FRAME_DATA:
-      rv = nghttp3_stream_write_data(stream, &data_eof, frent);
+      rv = nghttp3_stream_write_data(stream, &data_eof, &fr->data);
       if (rv != 0) {
         return rv;
       }
@@ -280,18 +280,24 @@ int nghttp3_stream_fill_outq(nghttp3_stream *stream) {
       }
       break;
     case NGHTTP3_FRAME_GOAWAY:
-      rv = nghttp3_stream_write_goaway(stream, frent);
+      rv = nghttp3_stream_write_goaway(stream, &fr->goaway);
       if (rv != 0) {
         return rv;
       }
       break;
     case NGHTTP3_FRAME_PRIORITY_UPDATE:
-      rv = nghttp3_stream_write_priority_update(stream, frent);
+      rv = nghttp3_stream_write_priority_update(stream, &fr->priority_update);
       if (rv != 0) {
         return rv;
       }
-      nghttp3_frame_priority_update_free(&frent->fr.priority_update,
-                                         stream->mem);
+      nghttp3_frame_priority_update_free(&fr->priority_update, stream->mem);
+      break;
+    case NGHTTP3_FRAME_ORIGIN:
+      rv = nghttp3_stream_write_origin(stream, &fr->origin);
+      if (rv != 0) {
+        return rv;
+      }
+
       break;
     default:
       /* TODO Not implemented */
@@ -325,51 +331,56 @@ int nghttp3_stream_write_stream_type(nghttp3_stream *stream) {
 }
 
 int nghttp3_stream_write_settings(nghttp3_stream *stream,
-                                  nghttp3_frame_entry *frent) {
+                                  const nghttp3_frame_settings *infr) {
   size_t len;
   int rv;
   nghttp3_buf *chunk;
   nghttp3_typed_buf tbuf;
-  struct {
-    nghttp3_frame_settings settings;
-    nghttp3_settings_entry iv[15];
-  } fr = {
-    .settings =
-      {
-        .hd =
-          {
-            .type = NGHTTP3_FRAME_SETTINGS,
-          },
-        .niv = 3,
-      },
-  };
-  nghttp3_settings_entry *iv;
-  nghttp3_settings *local_settings = frent->aux.settings.local_settings;
+  nghttp3_settings_entry ents[16];
+  nghttp3_frame_settings fr = *infr;
+  const nghttp3_settings *local_settings = infr->local_settings;
+  int64_t payloadlen;
 
-  iv = &fr.settings.iv[0];
+  if (local_settings) {
+    fr = (nghttp3_frame_settings){
+      .type = NGHTTP3_FRAME_SETTINGS,
+      .niv = 3,
+      .iv = ents,
+    };
 
-  iv[0].id = NGHTTP3_SETTINGS_ID_MAX_FIELD_SECTION_SIZE;
-  iv[0].value = local_settings->max_field_section_size;
-  iv[1].id = NGHTTP3_SETTINGS_ID_QPACK_MAX_TABLE_CAPACITY;
-  iv[1].value = local_settings->qpack_max_dtable_capacity;
-  iv[2].id = NGHTTP3_SETTINGS_ID_QPACK_BLOCKED_STREAMS;
-  iv[2].value = local_settings->qpack_blocked_streams;
+    ents[0] = (nghttp3_settings_entry){
+      .id = NGHTTP3_SETTINGS_ID_MAX_FIELD_SECTION_SIZE,
+      .value = local_settings->max_field_section_size,
+    };
+    ents[1] = (nghttp3_settings_entry){
+      .id = NGHTTP3_SETTINGS_ID_QPACK_MAX_TABLE_CAPACITY,
+      .value = local_settings->qpack_max_dtable_capacity,
+    };
+    ents[2] = (nghttp3_settings_entry){
+      .id = NGHTTP3_SETTINGS_ID_QPACK_BLOCKED_STREAMS,
+      .value = local_settings->qpack_blocked_streams,
+    };
 
-  if (local_settings->h3_datagram) {
-    iv[fr.settings.niv].id = NGHTTP3_SETTINGS_ID_H3_DATAGRAM;
-    iv[fr.settings.niv].value = 1;
+    if (local_settings->h3_datagram) {
+      ents[fr.niv] = (nghttp3_settings_entry){
+        .id = NGHTTP3_SETTINGS_ID_H3_DATAGRAM,
+        .value = 1,
+      };
 
-    ++fr.settings.niv;
+      ++fr.niv;
+    }
+
+    if (local_settings->enable_connect_protocol) {
+      ents[fr.niv] = (nghttp3_settings_entry){
+        .id = NGHTTP3_SETTINGS_ID_ENABLE_CONNECT_PROTOCOL,
+        .value = 1,
+      };
+
+      ++fr.niv;
+    }
   }
 
-  if (local_settings->enable_connect_protocol) {
-    iv[fr.settings.niv].id = NGHTTP3_SETTINGS_ID_ENABLE_CONNECT_PROTOCOL;
-    iv[fr.settings.niv].value = 1;
-
-    ++fr.settings.niv;
-  }
-
-  len = nghttp3_frame_write_settings_len(&fr.settings.hd.length, &fr.settings);
+  len = nghttp3_frame_write_settings_len(&payloadlen, &fr);
 
   rv = nghttp3_stream_ensure_chunk(stream, len);
   if (rv != 0) {
@@ -379,7 +390,7 @@ int nghttp3_stream_write_settings(nghttp3_stream *stream,
   chunk = nghttp3_stream_get_chunk(stream);
   nghttp3_typed_buf_shared_init(&tbuf, chunk);
 
-  chunk->last = nghttp3_frame_write_settings(chunk->last, &fr.settings);
+  chunk->last = nghttp3_frame_write_settings(chunk->last, &fr, payloadlen);
 
   tbuf.buf.last = chunk->last;
 
@@ -387,14 +398,14 @@ int nghttp3_stream_write_settings(nghttp3_stream *stream,
 }
 
 int nghttp3_stream_write_goaway(nghttp3_stream *stream,
-                                nghttp3_frame_entry *frent) {
-  nghttp3_frame_goaway *fr = &frent->fr.goaway;
+                                const nghttp3_frame_goaway *fr) {
   size_t len;
   int rv;
   nghttp3_buf *chunk;
   nghttp3_typed_buf tbuf;
+  int64_t payloadlen;
 
-  len = nghttp3_frame_write_goaway_len(&fr->hd.length, fr);
+  len = nghttp3_frame_write_goaway_len(&payloadlen, fr);
 
   rv = nghttp3_stream_ensure_chunk(stream, len);
   if (rv != 0) {
@@ -404,22 +415,22 @@ int nghttp3_stream_write_goaway(nghttp3_stream *stream,
   chunk = nghttp3_stream_get_chunk(stream);
   nghttp3_typed_buf_shared_init(&tbuf, chunk);
 
-  chunk->last = nghttp3_frame_write_goaway(chunk->last, fr);
+  chunk->last = nghttp3_frame_write_goaway(chunk->last, fr, payloadlen);
 
   tbuf.buf.last = chunk->last;
 
   return nghttp3_stream_outq_add(stream, &tbuf);
 }
 
-int nghttp3_stream_write_priority_update(nghttp3_stream *stream,
-                                         nghttp3_frame_entry *frent) {
-  nghttp3_frame_priority_update *fr = &frent->fr.priority_update;
+int nghttp3_stream_write_priority_update(
+  nghttp3_stream *stream, const nghttp3_frame_priority_update *fr) {
   size_t len;
   int rv;
   nghttp3_buf *chunk;
   nghttp3_typed_buf tbuf;
+  int64_t payloadlen;
 
-  len = nghttp3_frame_write_priority_update_len(&fr->hd.length, fr);
+  len = nghttp3_frame_write_priority_update_len(&payloadlen, fr);
 
   rv = nghttp3_stream_ensure_chunk(stream, len);
   if (rv != 0) {
@@ -429,16 +440,54 @@ int nghttp3_stream_write_priority_update(nghttp3_stream *stream,
   chunk = nghttp3_stream_get_chunk(stream);
   nghttp3_typed_buf_shared_init(&tbuf, chunk);
 
-  chunk->last = nghttp3_frame_write_priority_update(chunk->last, fr);
+  chunk->last =
+    nghttp3_frame_write_priority_update(chunk->last, fr, payloadlen);
 
   tbuf.buf.last = chunk->last;
 
   return nghttp3_stream_outq_add(stream, &tbuf);
 }
 
+int nghttp3_stream_write_origin(nghttp3_stream *stream,
+                                const nghttp3_frame_origin *fr) {
+  nghttp3_buf *chunk;
+  nghttp3_buf buf;
+  nghttp3_typed_buf tbuf;
+  int rv;
+
+  rv = nghttp3_stream_ensure_chunk(
+    stream, nghttp3_frame_write_hd_len(fr->type, (int64_t)fr->origin_list.len));
+  if (rv != 0) {
+    return rv;
+  }
+
+  chunk = nghttp3_stream_get_chunk(stream);
+  nghttp3_typed_buf_shared_init(&tbuf, chunk);
+
+  chunk->last =
+    nghttp3_frame_write_hd(chunk->last, fr->type, (int64_t)fr->origin_list.len);
+
+  tbuf.buf.last = chunk->last;
+
+  rv = nghttp3_stream_outq_add(stream, &tbuf);
+  if (rv != 0) {
+    return rv;
+  }
+
+  if (fr->origin_list.len == 0) {
+    return 0;
+  }
+
+  nghttp3_buf_wrap_init(&buf, (uint8_t *)fr->origin_list.base,
+                        fr->origin_list.len);
+  buf.last = buf.end;
+  nghttp3_typed_buf_init(&tbuf, &buf, NGHTTP3_BUF_TYPE_ALIEN_NO_ACK);
+
+  return nghttp3_stream_outq_add(stream, &tbuf);
+}
+
 int nghttp3_stream_write_headers(nghttp3_stream *stream,
-                                 nghttp3_frame_entry *frent) {
-  nghttp3_frame_headers *fr = &frent->fr.headers;
+                                 const nghttp3_frame_headers *fr) {
   nghttp3_conn *conn = stream->conn;
 
   assert(conn);
@@ -459,9 +508,9 @@ int nghttp3_stream_write_header_block(nghttp3_stream *stream,
   size_t len;
   nghttp3_buf *chunk;
   nghttp3_typed_buf tbuf;
-  nghttp3_frame_hd hd;
   uint8_t raw_pbuf[16];
   size_t pbuflen, rbuflen, ebuflen;
+  int64_t payloadlen;
 
   nghttp3_buf_wrap_init(&pbuf, raw_pbuf, sizeof(raw_pbuf));
 
@@ -475,10 +524,9 @@ int nghttp3_stream_write_header_block(nghttp3_stream *stream,
   rbuflen = nghttp3_buf_len(rbuf);
   ebuflen = nghttp3_buf_len(ebuf);
 
-  hd.type = frame_type;
-  hd.length = (int64_t)(pbuflen + rbuflen);
+  payloadlen = (int64_t)(pbuflen + rbuflen);
 
-  len = nghttp3_frame_write_hd_len(&hd) + pbuflen;
+  len = nghttp3_frame_write_hd_len(frame_type, payloadlen) + pbuflen;
 
   if (rbuflen <= NGHTTP3_STREAM_MAX_COPY_THRES) {
     len += rbuflen;
@@ -492,7 +540,7 @@ int nghttp3_stream_write_header_block(nghttp3_stream *stream,
   chunk = nghttp3_stream_get_chunk(stream);
   nghttp3_typed_buf_shared_init(&tbuf, chunk);
 
-  chunk->last = nghttp3_frame_write_hd(chunk->last, &hd);
+  chunk->last = nghttp3_frame_write_hd(chunk->last, frame_type, payloadlen);
 
   chunk->last = nghttp3_cpymem(chunk->last, pbuf.pos, pbuflen);
   nghttp3_buf_init(&pbuf);
@@ -560,17 +608,16 @@ int nghttp3_stream_write_header_block(nghttp3_stream *stream,
 }
 
 int nghttp3_stream_write_data(nghttp3_stream *stream, int *peof,
-                              nghttp3_frame_entry *frent) {
+                              const nghttp3_frame_data *fr) {
   int rv;
   size_t len;
   nghttp3_typed_buf tbuf;
   nghttp3_buf buf;
   nghttp3_buf *chunk;
-  nghttp3_read_data_callback read_data = frent->aux.data.dr.read_data;
+  nghttp3_read_data_callback read_data = fr->dr.read_data;
   nghttp3_conn *conn = stream->conn;
   int64_t datalen;
   uint32_t flags = 0;
-  nghttp3_frame_hd hd;
   nghttp3_vec vec[8];
   nghttp3_vec *v;
   nghttp3_ssize sveccnt;
@@ -624,10 +671,7 @@ int nghttp3_stream_write_data(nghttp3_stream *stream, int *peof,
     }
   }
 
-  hd.type = NGHTTP3_FRAME_DATA;
-  hd.length = datalen;
-
-  len = nghttp3_frame_write_hd_len(&hd);
+  len = nghttp3_frame_write_hd_len(NGHTTP3_FRAME_DATA, datalen);
 
   rv = nghttp3_stream_ensure_chunk(stream, len);
   if (rv != 0) {
@@ -637,7 +681,8 @@ int nghttp3_stream_write_data(nghttp3_stream *stream, int *peof,
   chunk = nghttp3_stream_get_chunk(stream);
   nghttp3_typed_buf_shared_init(&tbuf, chunk);
 
-  chunk->last = nghttp3_frame_write_hd(chunk->last, &hd);
+  chunk->last =
+    nghttp3_frame_write_hd(chunk->last, NGHTTP3_FRAME_DATA, datalen);
 
   tbuf.buf.last = chunk->last;
 
@@ -646,19 +691,19 @@ int nghttp3_stream_write_data(nghttp3_stream *stream, int *peof,
     return rv;
   }
 
-  if (datalen) {
-    for (i = 0; i < (size_t)sveccnt; ++i) {
-      v = &vec[i];
-      if (v->len == 0) {
-        continue;
-      }
-      nghttp3_buf_wrap_init(&buf, v->base, v->len);
-      buf.last = buf.end;
-      nghttp3_typed_buf_init(&tbuf, &buf, NGHTTP3_BUF_TYPE_ALIEN);
-      rv = nghttp3_stream_outq_add(stream, &tbuf);
-      if (rv != 0) {
-        return rv;
-      }
+  assert(datalen);
+
+  for (i = 0; i < (size_t)sveccnt; ++i) {
+    v = &vec[i];
+    if (v->len == 0) {
+      continue;
+    }
+    nghttp3_buf_wrap_init(&buf, v->base, v->len);
+    buf.last = buf.end;
+    nghttp3_typed_buf_init(&tbuf, &buf, NGHTTP3_BUF_TYPE_ALIEN);
+    rv = nghttp3_stream_outq_add(stream, &tbuf);
+    if (rv != 0) {
+      return rv;
     }
   }
 
@@ -888,6 +933,7 @@ static void stream_pop_outq_entry(nghttp3_stream *stream,
     nghttp3_buf_free(&tbuf->buf, stream->mem);
     break;
   case NGHTTP3_BUF_TYPE_ALIEN:
+  case NGHTTP3_BUF_TYPE_ALIEN_NO_ACK:
     break;
   case NGHTTP3_BUF_TYPE_SHARED:
     assert(nghttp3_ringbuf_len(chunks));
@@ -899,7 +945,7 @@ static void stream_pop_outq_entry(nghttp3_stream *stream,
     if (chunk->last == tbuf->buf.last) {
       if (nghttp3_buf_cap(chunk) == NGHTTP3_STREAM_MIN_CHUNK_SIZE) {
         nghttp3_objalloc_chunk_release(stream->out_chunk_objalloc,
-                                       (nghttp3_chunk *)(void *)chunk->begin);
+                                       (void *)chunk->begin);
       } else {
         nghttp3_buf_free(chunk, stream->mem);
       }
@@ -908,7 +954,7 @@ static void stream_pop_outq_entry(nghttp3_stream *stream,
     break;
   default:
     nghttp3_unreachable();
-  };
+  }
 
   nghttp3_ringbuf_pop_front(&stream->outq);
 }
